//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      GUI/Model/Descriptor/SelectionProperty.h
//! @brief     Defines class SelectionProperty
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2021
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#ifndef BORNAGAIN_GUI_MODEL_DESCRIPTOR_SELECTIONPROPERTY_H
#define BORNAGAIN_GUI_MODEL_DESCRIPTOR_SELECTIONPROPERTY_H

#include "GUI/Model/Descriptor/SelectionDescriptor.h"
#include "GUI/Support/XML/Streamer.h"

//! Abstract base class for SelectionProperty to ease referencing.
class AbstractSelectionProperty {
public:
    virtual ~AbstractSelectionProperty() = default;

    //! Set currently selected option
    virtual void setCurrentIndex(int newIndex) = 0;

    //! Get currently selected option
    virtual int currentIndex() const = 0;
};

//! Class for representing a selection, its attributes and its accessors.
//!
//! A "selection" in this context is a class instance out of a range of possible class instances.
//!
//! Example:
//! A distribution can be represented by DistributionGateItem, DistributionLorentzItem and more. The
//! common base class is DistributionItem. To store the currently distribution, you may use a member
//! of DistributionItem* which than holds an instance of the currently selected distribution (e.g.
//! DistributionGateItem*). To select between distributions, you need a combo box filled with the
//! distribution names, a label for the combo box, and so on.
//!
//! This class provides many of these attributes and functionalities:
//! * A pointer to the common base class (given as the template argument T). The lifetime of this
//! pointer is handled in here.
//! * label: a label of e.g. a spin box
//! * tooltip: tooltip for e.g. a spin box
//! * persistent tag: a name to serialize this property. Do not change this string, since it is
//!   serialized!
//! * A list of available options (e.g. the names of the distributions)
//! * A descriptor which provides some of the given information, as well as getters/setters.
//!
//! For the initialization of a SelectionProperty there are helper methods. They are not necessarily
//! to be called if they are not convenient. They are only there to help initialization e.g by using
//! a catalog. If no catalog exists for the given case, the initialization can be done by any means.
//!
//! This means especially that this class is **not** relying on the existence of a related catalog
//! class - a catalog helps but is not mandatory.
//!
//! \sa SelectionDescriptor
//!
template <typename Catalog>
class SelectionProperty : public AbstractSelectionProperty{
public:
    using CatalogedType = typename Catalog::CatalogedType;

    //! Initialize by means of a catalog class and optional creation arguments.
    //!
    //! The current selection will be initialized with the first type in the catalog types. The
    //! optional arguments are the arguments which may be necessary for the creation method in the
    //! catalog.
    template <typename... ArgsForCreation>
    void init(const QString& label, const QString& tooltip, ArgsForCreation... argsForCreation)
    {
        initFieldsAndSetter(label, tooltip, argsForCreation...);
        setCurrentIndex(0);

        initDescriptor(label, tooltip, argsForCreation...);
        m_descriptor.setCurrentIndex(0);
    }

    //! Initialize by means of a catalog class and an initializer function.
    //!
    //! The current selection will be initialized with the first type in the catalog types.
    //! Each newly created item can be initialized with the given initializer method. Note that the
    //! item creation will take place also after a call to this method, so take care that the given
    //! initializer stays valid throughout the lifetime of this SelectionProperty instance.
    //!
    //! The initializer function takes two arguments: The first is the new item, the second is the
    //! old one (if present; can be null). This is intended to maybe copy values from the old to the
    //! new selection. The old item also can be ignored, always according to the current
    //! needs.
    void initWithInitializer(
        const QString& label, const QString& tooltip,
        std::function<void(CatalogedType* newItem, const CatalogedType* oldItem)> initializer)
    {
        m_initializer = initializer;

        initFieldsAndSetter(label, tooltip);
        setCurrentIndex(0);

        initDescriptor(label, tooltip);
        m_descriptor.setCurrentIndex(0);
    }

    //! Initialize by means of a catalog class, a subsection of allowed types and an initializer
    //! function.
    //!
    //! Same as before, but only a subset of types available in a catalog shall be used.
    //! The current selection will be initialized with the first type in the given types subset.
    //! Each newly created item can be initialized with the given initializer method. Note that the
    //! item creation will take place also after a call to this method, so take care that the given
    //! initializer stays valid throughout the lifetime of this SelectionProperty instance.
    //!
    //! The initializer function takes two arguments: The first is the new item, the second is the
    //! old one (if present; can be null). This is intended to maybe copy values from the old to the
    //! new selection. The old item also can be ignored, always according to the current
    //! needs.
    void initWithInitializer(
        const QString& label, const QString& tooltip, const QVector<typename Catalog::Type>& types,
        std::function<void(CatalogedType* newItem, const CatalogedType* oldItem)> initializer)
    {
        m_initializer = initializer;
        m_types = types;

        initFieldsAndSetter(label, tooltip);
        setCurrentIndex(0);

        initDescriptor(label, tooltip);
        m_descriptor.setCurrentIndex(0);
    }

    //! Cast to a descriptor working on this property.
    operator SelectionDescriptor<CatalogedType*>() const { return m_descriptor; }

    //! Direct access to the stored pointer
    CatalogedType* operator->() const { return m_p.get(); }

    //! Direct access to the stored pointer
    CatalogedType* currentItem() const {return m_p.get();}

    //! Directly set the new item.
    void setCurrentItem(CatalogedType* t, bool callInitializer = false)
    {
        if (callInitializer && m_initializer)
            m_initializer(t, m_p.get());
        m_p.reset(t);
    }

    //! Directly set the new item.
    template <typename S, typename... ArgsForConstructor>
    S* setCurrentItem(ArgsForConstructor... argsForConstructor)
    {
        S* s = new S(argsForConstructor...);
        if (s != nullptr && m_initializer)
            m_initializer(s, m_p.get());
        m_p.reset(s);
        return s;
    }

    template <typename... ArgsForCreation>
    void rwSelected(Streamer& s, const QString& tag, ArgsForCreation... argsForCreation)
    {
        if (s.xmlWriter()) {
            s.write<Catalog>(tag, m_p.get());
        } else if (s.xmlReader()) {
            CatalogedType* p = nullptr;
            s.read<Catalog>(tag, p, argsForCreation...);
            m_p.reset(p);
        }
    }

    QString label() const {return m_label;}
    QString tooltip() const {return m_tooltip;}
    QStringList options() const {return m_options;}

    void setCurrentIndex(int newIndex) override { currentIndexSetter(newIndex); }
    int currentIndex() const override { return m_types.indexOf(Catalog::type(m_p.get())); }

    SelectionDescriptor<CatalogedType*>
        m_descriptor; //!< descriptor, holding attributes like label, tooltip

private:
    template <typename... ArgsForCreation>
    void initFieldsAndSetter(const QString& label, const QString& tooltip, ArgsForCreation... argsForCreation) {
        m_label = label;
        m_tooltip = tooltip;
        m_options.clear();
        for (const auto type : m_types)
            m_options << Catalog::uiInfo(type).menuEntry;

        currentIndexSetter = [=](int current) {
            auto* p = Catalog::create(m_types[current], argsForCreation...);
            if (m_initializer)
                m_initializer(p, m_p.get());
            m_p.reset(p);
        };
    }

    template <typename... ArgsForCreation>
    void initDescriptor(const QString& label, const QString& tooltip,
                        ArgsForCreation... argsForCreation)

    {
        auto setCurrentIndex = [=](int current) {
            auto* p = Catalog::create(m_types[current], argsForCreation...);
            if (m_initializer)
                m_initializer(p, m_p.get());
            m_p.reset(p);
        };

        m_descriptor.label = label;
        m_descriptor.tooltip = tooltip;
        for (const auto type : m_types)
            m_descriptor.options << Catalog::uiInfo(type).menuEntry;
        m_descriptor.currentIndexSetter = setCurrentIndex;
        m_descriptor.currentIndexGetter = [=]() {
            return m_types.indexOf(Catalog::type(m_p.get()));
        };
        m_descriptor.currentItem = [=] { return m_p.get(); };
    }

private:
    std::unique_ptr<CatalogedType> m_p; //!< Current selection

    QString m_label;       //!< A label text (short, no trailing colon)
    QString m_tooltip;     //!< Tooltip text
    QStringList m_options; //!< List of options, usually presented as combo entries
    QVector<typename Catalog::Type> m_types = Catalog::types();

    //! Reimplementable function to set currently selected option.
    std::function<void(int)> currentIndexSetter = nullptr;

    //! initializer function. Can be empty.
    //! The first argument is the new item, the second is the old one if present; can be null.
    //! This is intended to maybe copy values from the old to the new selection. oldItem also can be
    //! ignored if not relevant,
    std::function<void(CatalogedType* newItem, const CatalogedType* oldItem)> m_initializer;
};

#endif // BORNAGAIN_GUI_MODEL_DESCRIPTOR_SELECTIONPROPERTY_H
