//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      GUI/Model/Device/SpecularBeamInclinationItem.cpp
//! @brief     Implements class SpecularBeamInclinationItem
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2018
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#include "GUI/Model/Device/SpecularBeamInclinationItem.h"
#include "Base/Axis/IAxis.h"
#include "Base/Const/Units.h"
#include "GUI/Model/CatDevice/DistributionItemCatalog.h"
#include "GUI/Model/Device/PointwiseAxisItem.h"
#include "GUI/Support/XML/Serialize.h"
#include "GUI/Support/XML/UtilXML.h"

namespace {

namespace Tag {

const QString UniformAxis("UniformAxis");
const QString PointwiseAxis("PointwiseAxis");
const QString BeamInclinationDistribution("BeamInclinationDistribution");

} // namespace Tag


void setAxisPresentationDefaults(BasicAxisItem* axisItem)
{
    ASSERT(axisItem);
    axisItem->setTitle("alpha_i");

    if (!dynamic_cast<PointwiseAxisItem*>(axisItem)) {
        axisItem->setMin(0.0);
        axisItem->setMax(3.0);
        axisItem->setBinCount(500);
    }
}

void initDistribution(DistributionItem* newDistribution,
                      const DistributionItem* currentDistribution)
{
    newDistribution->setLimits(RealLimits::limited(-90.0, 90.0));
    newDistribution->setUnit(Unit::degree);

    double meanVal = 0.0;
    if (auto* cd = dynamic_cast<const SymmetricResolutionItem*>(currentDistribution))
        meanVal = cd->mean();

    if (auto* nd = dynamic_cast<SymmetricResolutionItem*>(newDistribution)) {
        nd->setMean(meanVal);
        nd->setMeanDecimals(3);
    }

    newDistribution->initDistribution(meanVal);
}

} // namespace

SpecularBeamInclinationItem::SpecularBeamInclinationItem(const InstrumentItem* owningInstrument)
    : BeamDistributionItem()
    , m_owningInstrument(owningInstrument)
{
    m_distribution.initWithInitializer("Distribution", "",
                                       DistributionItemCatalog::symmetricTypes(), initDistribution);

    m_uniformAlphaAxis.reset(new BasicAxisItem());
    setAxisPresentationDefaults(m_uniformAlphaAxis.get());
    m_currentAxisIsUniformAxis = true;
}

void SpecularBeamInclinationItem::serialize(Streamer& s)
{
    s.assertVersion(0);
    m_distribution.rwSelected(s, Tag::BeamInclinationDistribution);
    Serialize::rwValue(s, "currentAxisIsUniformAxis", m_currentAxisIsUniformAxis);

    // r/w axes
    if (auto* w = s.xmlWriter())
        writeTo(w);

    if (auto* r = s.xmlReader()) {
        readFrom(r);
        m_distribution->setLimits(RealLimits::limited(-90.0, 90.0));
        m_distribution->setUnit(Unit::degree);
        if (m_pointwiseAlphaAxis)
            m_pointwiseAlphaAxis->setOwningInstrument(m_owningInstrument);
    }
}

void SpecularBeamInclinationItem::writeTo(QXmlStreamWriter* writer) const
{
    // uniform axis
    if (m_uniformAlphaAxis) {
        writer->writeStartElement(Tag::UniformAxis);
        m_uniformAlphaAxis->writeTo(writer);
        writer->writeEndElement();
    }

    // pointwise axis
    if (m_pointwiseAlphaAxis) {
        writer->writeStartElement(Tag::PointwiseAxis);
        m_pointwiseAlphaAxis->writeTo(writer);
        writer->writeEndElement();
    }
}

void SpecularBeamInclinationItem::readFrom(QXmlStreamReader* reader)
{
    while (reader->readNextStartElement()) {

        // uniform axis
        if (reader->name() == Tag::UniformAxis) {
            m_uniformAlphaAxis = std::make_unique<BasicAxisItem>();
            setAxisPresentationDefaults(m_uniformAlphaAxis.get());
            m_uniformAlphaAxis->readFrom(reader);
            GUI::Session::XML::gotoEndElementOfTag(reader, Tag::UniformAxis);

            // pointwise axis
        } else if (reader->name() == Tag::PointwiseAxis) {
            m_pointwiseAlphaAxis = std::make_unique<PointwiseAxisItem>();
            setAxisPresentationDefaults(m_pointwiseAlphaAxis.get());
            m_pointwiseAlphaAxis->readFrom(reader);
            GUI::Session::XML::gotoEndElementOfTag(reader, Tag::PointwiseAxis);

        } else
            reader->skipCurrentElement();
    }
}

double SpecularBeamInclinationItem::scaleFactor() const
{
    return Units::deg;
}

BasicAxisItem* SpecularBeamInclinationItem::alphaAxis() const
{
    return m_currentAxisIsUniformAxis ? m_uniformAlphaAxis.get() : m_pointwiseAlphaAxis.get();
}

bool SpecularBeamInclinationItem::pointwiseAlphaAxisDefined() const
{
    return m_pointwiseAlphaAxis.get() != nullptr;
}

bool SpecularBeamInclinationItem::pointwiseAlphaAxisSelected() const
{
    return !m_currentAxisIsUniformAxis;
}

bool SpecularBeamInclinationItem::uniformAlphaAxisSelected() const
{
    return m_currentAxisIsUniformAxis;
}

void SpecularBeamInclinationItem::selectUniformAxis()
{
    m_currentAxisIsUniformAxis = true;
}

void SpecularBeamInclinationItem::selectPointwiseAxis()
{
    ASSERT(pointwiseAlphaAxisDefined());
    m_currentAxisIsUniformAxis = false;
}

void SpecularBeamInclinationItem::initUniformAxis(const IAxis& axis)
{
    m_uniformAlphaAxis->setBinCount(static_cast<int>(axis.size()));
}

void SpecularBeamInclinationItem::initPointwiseAxis(const IAxis& axis, QString units)
{
    if (!m_pointwiseAlphaAxis) {
        m_pointwiseAlphaAxis.reset(new PointwiseAxisItem());
        m_pointwiseAlphaAxis->setOwningInstrument(m_owningInstrument);
        setAxisPresentationDefaults(m_pointwiseAlphaAxis.get());
    }

    m_pointwiseAlphaAxis->init(axis, units);
    m_pointwiseAlphaAxis->updateIndicators();
}
