//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      GUI/Model/Job/FitSuiteItem.cpp
//! @brief     Implements class FitSuiteItem
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2018
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#include "GUI/Model/Job/FitSuiteItem.h"
#include "GUI/Model/Data/SessionData.h"
#include "GUI/Model/Job/FitParameterContainerItem.h"
#include "GUI/Model/Job/MinimizerItem.h"
#include "GUI/Model/Project/ProjectDocument.h"
#include "GUI/Support/XML/UtilXML.h"
#include "GUI/Util/Error.h"

namespace {

namespace Tag {

const QString UpdateInterval("UpdateInterval");
const QString IterationsCount("IterationsCount");
const QString Chi2("Chi2");

const QString MinimizerContainer("MinimizerContainer");
const QString FitParameterContainer("FitParameterContainer");

} // namespace Tag

namespace Attrib {

const QString value("value");

} // namespace Attrib

} // namespace

FitSuiteItem::FitSuiteItem()
    : m_updateInterval(10)
    , m_iterCount(0)
    , m_chi2(0.0)
{
    setObjectName("FitSuite");
}

FitParameterContainerItem* FitSuiteItem::createFitParametersContainer()
{
    if (m_fitContainer)
        throw Error("FitSuiteItem::createFitParametersContainer() -> Error. Attempt to create "
                    "a second FitParameterContainer.");

    m_fitContainer = std::make_unique<FitParameterContainerItem>();
    return m_fitContainer.get();
}

FitParameterContainerItem* FitSuiteItem::fitParameterContainerItem()
{
    return m_fitContainer.get();
}

MinimizerContainerItem* FitSuiteItem::createMinimizerContainer()
{
    if (m_minimizerContainer)
        throw Error("FitSuiteItem::createMinimizerContainer() -> Error. Attempt to create "
                    "a second MinimizerContainer.");

    m_minimizerContainer = std::make_unique<MinimizerContainerItem>();
    return m_minimizerContainer.get();
}

MinimizerContainerItem* FitSuiteItem::minimizerContainerItem()
{
    return m_minimizerContainer.get();
}

int FitSuiteItem::updateInterval() const
{
    return m_updateInterval;
}

void FitSuiteItem::setUpdateInterval(int interval)
{
    m_updateInterval = interval;
    emit updateIntervalChanged(interval);
    gSessionData->projectDocument.value()->setModified();
}

int FitSuiteItem::iterationCount() const
{
    return m_iterCount;
}

void FitSuiteItem::setIterationCount(int count)
{
    m_iterCount = count;
    emit iterationCountChanged(count);
}

double FitSuiteItem::chi2() const
{
    return m_chi2;
}

void FitSuiteItem::setChi2(const double chi2)
{
    m_chi2 = chi2;
}

void FitSuiteItem::writeTo(QXmlStreamWriter* writer) const
{
    // update interval
    writer->writeStartElement(Tag::UpdateInterval);
    GUI::Session::XML::writeAttribute(writer, Attrib::value, m_updateInterval);
    writer->writeEndElement();

    // iterations count
    writer->writeStartElement(Tag::IterationsCount);
    GUI::Session::XML::writeAttribute(writer, Attrib::value, m_iterCount);
    writer->writeEndElement();

    // chi2
    writer->writeStartElement(Tag::Chi2);
    GUI::Session::XML::writeAttribute(writer, Attrib::value, m_chi2);
    writer->writeEndElement();

    // minimizer container
    if (m_minimizerContainer) {
        writer->writeStartElement(Tag::MinimizerContainer);
        m_minimizerContainer->writeTo(writer);
        writer->writeEndElement();
    }

    // fit parameter container
    if (m_fitContainer) {
        writer->writeStartElement(Tag::FitParameterContainer);
        m_fitContainer->writeTo(writer);
        writer->writeEndElement();
    }
}

void FitSuiteItem::readFrom(QXmlStreamReader* reader)
{
    while (reader->readNextStartElement()) {

        // update interval
        if (reader->name() == Tag::UpdateInterval) {
            GUI::Session::XML::readAttribute(reader, Attrib::value, &m_updateInterval);
            GUI::Session::XML::gotoEndElementOfTag(reader, Tag::UpdateInterval);

            // iterations count
        } else if (reader->name() == Tag::IterationsCount) {
            GUI::Session::XML::readAttribute(reader, Attrib::value, &m_iterCount);
            GUI::Session::XML::gotoEndElementOfTag(reader, Tag::IterationsCount);

            // chi2
        } else if (reader->name() == Tag::Chi2) {
            GUI::Session::XML::readAttribute(reader, Attrib::value, &m_chi2);
            GUI::Session::XML::gotoEndElementOfTag(reader, Tag::Chi2);

            // minimizer container
        } else if (reader->name() == Tag::MinimizerContainer) {
            createMinimizerContainer()->readFrom(reader);
            GUI::Session::XML::gotoEndElementOfTag(reader, Tag::MinimizerContainer);

            // fit parameter container
        } else if (reader->name() == Tag::FitParameterContainer) {
            createFitParametersContainer()->readFrom(reader);
            GUI::Session::XML::gotoEndElementOfTag(reader, Tag::FitParameterContainer);

        } else
            reader->skipCurrentElement();
    }
}
