//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      GUI/Model/Job/MinimizerItem.cpp
//! @brief     Implements MinimizerItem class
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2018
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#include "GUI/Model/Job/MinimizerItem.h"
#include "Base/Util/Assert.h"
#include "Fit/Adapter/GSLLevenbergMarquardtMinimizer.h"
#include "Fit/Adapter/GSLMultiMinimizer.h"
#include "Fit/Adapter/GeneticMinimizer.h"
#include "Fit/Adapter/Minuit2Minimizer.h"
#include "Fit/Adapter/SimAnMinimizer.h"
#include "GUI/Model/CatJob/MinimizerItemCatalog.h"
#include "GUI/Support/XML/UtilXML.h"
#include "Sim/Fitting/ObjectiveMetric.h"
#include "Sim/Fitting/ObjectiveMetricUtils.h"

namespace {

namespace Tag {

const QString Tolerance("Tolerance");
const QString MaxIterations("MaxIterations");
const QString IterationsAtTemp("IterationsAtTemp");
const QString StepSize("StepSize");
const QString BoltzmannK("BoltzmannK");
const QString Tinit("Tinit");
const QString BoltzmannMu("BoltzmannMu");
const QString Tmin("Tmin");
const QString PopulationSize("PopulationSize");
const QString RandomSeed("RandomSeed");
const QString Algorithm("Algorithm");
const QString Strategy("Strategy");
const QString ErrorDef("ErrorDef");
const QString Precision("Precision");
const QString MaxFuncCalls("MaxFuncCalls");
const QString Minimizer("Minimizer");
const QString Metric("Metric");
const QString Norm("Norm");
const QString MinuitMinimizer("MinuitMinimizer");
const QString GSLMultiMinimizer("GSLMultiMinimizer");
const QString GeneticMinimizer("GeneticMinimizer");
const QString SimAnMinimizer("SimAnMinimizer");
const QString GSLLMAMinimizer("GSLLMAMinimizer");

} // namespace Tag

namespace Attrib {

const QString value("value");

} // namespace Attrib

const QMap<QString, MinimizerType> minimizer_names_map = {
    {QString::fromStdString(MinimizerInfo::buildMinuit2Info().name()), MinimizerType::Minuit2},
    {QString::fromStdString(MinimizerInfo::buildGSLMultiMinInfo().name()),
     MinimizerType::GSLMultiMin},
    {QString::fromStdString(MinimizerInfo::buildGSLLMAInfo().name()), MinimizerType::GSLLMA},
    {QString::fromStdString(MinimizerInfo::buildGSLSimAnInfo().name()), MinimizerType::GSLSimAn},
    {QString::fromStdString(MinimizerInfo::buildGeneticInfo().name()), MinimizerType::Genetic}};

} // namespace

QString minimizerTypeToName(MinimizerType type)
{
    ASSERT(minimizer_names_map.values().contains(type));
    return minimizer_names_map.key(type);
}

namespace {

QMap<QString /*algorithm*/, QString /*minimizer*/> algorithm_minimizer_map;
void add_algorithm_from_minimizer_to_list_and_map(MinimizerType minimizer_type,
                                                  QStringList& common_algorithms_list,
                                                  QStringList& common_algorithms_descriptions)
{
    QString minimizer_name = minimizerTypeToName(minimizer_type);
    QStringList algorithms = MinimizerItemCatalog::algorithmNames(minimizer_name);
    QStringList descriptions = MinimizerItemCatalog::algorithmDescriptions(minimizer_name);
    ASSERT(algorithms.size() == descriptions.size())

    for (int i = 0; i < algorithms.size(); i++) {
        common_algorithms_list.append(algorithms[i]);
        common_algorithms_descriptions.append(descriptions[i]);
        algorithm_minimizer_map.insert(algorithms[i], minimizer_name);
    }
}
void create_algorithm_list_and_map(QString& default_common_algorithm, QStringList& algorithms_list,
                                   QStringList& algorithms_descriptions)
{
    // Group headers may not directly correspond to minimizers, so for them we don't use
    // descriptions from MinimizerInfo

    algorithms_list.clear();
    algorithms_descriptions.clear();

    // group 1
    // group header. Cannot be selected and have no mapped minimizer
    algorithms_list.append("Local optimization");
    algorithms_descriptions.append("");

    add_algorithm_from_minimizer_to_list_and_map(MinimizerType::Minuit2, algorithms_list,
                                                 algorithms_descriptions);

    add_algorithm_from_minimizer_to_list_and_map(MinimizerType::GSLMultiMin, algorithms_list,
                                                 algorithms_descriptions);

    add_algorithm_from_minimizer_to_list_and_map(MinimizerType::GSLLMA, algorithms_list,
                                                 algorithms_descriptions);

    // group 2
    algorithms_list.append("Global optimization");
    algorithms_descriptions.append("");

    // move "Scan" algorithm from the local group to the global
    qsizetype scan_index = algorithms_list.indexOf("Scan");
    ASSERT(scan_index >= 0);
    QString name = algorithms_list.takeAt(scan_index);
    QString descr = algorithms_descriptions.takeAt(scan_index);
    algorithms_list.append(name);
    algorithms_descriptions.append(descr);

    // other algorithms
    add_algorithm_from_minimizer_to_list_and_map(MinimizerType::Genetic, algorithms_list,
                                                 algorithms_descriptions);

    add_algorithm_from_minimizer_to_list_and_map(MinimizerType::GSLSimAn, algorithms_list,
                                                 algorithms_descriptions);

    default_common_algorithm = algorithms_list[1]; // after the first header: Minuit2 Migrad
}

} // namespace

MinimizerContainerItem::MinimizerContainerItem()
    : m_MinuitMinimizer(std::make_unique<MinuitMinimizerItem>())
    , m_GSLMultiMinimizer(std::make_unique<GSLMultiMinimizerItem>())
    , m_GeneticMinimizer(std::make_unique<GeneticMinimizerItem>())
    , m_SimAnMinimizer(std::make_unique<SimAnMinimizerItem>())
    , m_GSLLMAMinimizer(std::make_unique<GSLLMAMinimizerItem>())
{
    QString default_common_algorithm;
    QStringList common_algorithms_list;
    QStringList common_algorithms_descriptions;
    create_algorithm_list_and_map(default_common_algorithm, common_algorithms_list,
                                  common_algorithms_descriptions);
    m_algorithm = ComboProperty::fromList(common_algorithms_list, default_common_algorithm);
    m_algorithm.setToolTips(common_algorithms_descriptions);

    QString default_minimizer = minimizerTypeToName(MinimizerType::Minuit2);
    m_minimizer = ComboProperty::fromList(minimizer_names_map.keys(), default_minimizer);
    m_metric = ComboProperty::fromStdVec(ObjectiveMetricUtils::metricNames(),
                                         ObjectiveMetricUtils::defaultMetricName());
    m_norm = ComboProperty::fromStdVec(ObjectiveMetricUtils::normNames(),
                                       ObjectiveMetricUtils::defaultNormName());
}

MinimizerContainerItem::~MinimizerContainerItem() = default;

MinuitMinimizerItem* MinimizerContainerItem::minimizerItemMinuit() const
{
    return m_MinuitMinimizer.get();
}

GSLMultiMinimizerItem* MinimizerContainerItem::minimizerItemGSLMulti() const
{
    return m_GSLMultiMinimizer.get();
}

GeneticMinimizerItem* MinimizerContainerItem::minimizerItemGenetic() const
{
    return m_GeneticMinimizer.get();
}

SimAnMinimizerItem* MinimizerContainerItem::minimizerItemSimAn() const
{
    return m_SimAnMinimizer.get();
}

GSLLMAMinimizerItem* MinimizerContainerItem::minimizerItemGSLLMA() const
{
    return m_GSLLMAMinimizer.get();
}

MinimizerItem* MinimizerContainerItem::currentMinimizerItem() const
{
    if (currentMinimizer() == minimizerTypeToName(MinimizerType::Minuit2))
        return minimizerItemMinuit();

    else if (currentMinimizer() == minimizerTypeToName(MinimizerType::GSLMultiMin))
        return minimizerItemGSLMulti();

    else if (currentMinimizer() == minimizerTypeToName(MinimizerType::Genetic))
        return minimizerItemGenetic();

    else if (currentMinimizer() == minimizerTypeToName(MinimizerType::GSLSimAn))
        return minimizerItemSimAn();

    else if (currentMinimizer() == minimizerTypeToName(MinimizerType::GSLLMA))
        return minimizerItemGSLLMA();

    else
        ASSERT(0);

    return nullptr;
}

QString MinimizerContainerItem::currentMinimizer() const
{
    return m_minimizer.currentValue();
}

void MinimizerContainerItem::setCurrentMinimizer(const QString& name)
{
    m_minimizer.setCurrentValue(name);
}

bool MinimizerContainerItem::algorithmHasMinimizer(const QString& name)
{
    return algorithm_minimizer_map.contains(name);
}

void MinimizerContainerItem::setCurrentCommonAlgorithm(const QString& name)
{
    m_algorithm.setCurrentValue(name);

    ASSERT(algorithmHasMinimizer(name));
    setCurrentMinimizer(algorithm_minimizer_map.value(name));
    applyAlgorithmToMinimizer(name);
}

ComboProperty MinimizerContainerItem::commonAlgorithmCombo() const
{
    return m_algorithm;
}

void MinimizerContainerItem::applyAlgorithmToMinimizer(const QString& name)
{
    // Minuit2
    if (currentMinimizer() == minimizerTypeToName(MinimizerType::Minuit2))
        minimizerItemMinuit()->setCurrentAlgorithm(name);

    // GSL MultiMin
    if (currentMinimizer() == minimizerTypeToName(MinimizerType::GSLMultiMin))
        minimizerItemGSLMulti()->setCurrentAlgorithm(name);

    // TMVA Genetic
    // do nothing

    // GSL Simulated Annealing
    // do nothing

    // GSL Levenberg-Marquardt
    // do nothing
}

QString MinimizerContainerItem::currentObjectiveMetric() const
{
    return m_metric.currentValue();
}

void MinimizerContainerItem::setCurrentObjectiveMetric(const QString& name)
{
    m_metric.setCurrentValue(name);
}

ComboProperty MinimizerContainerItem::objectiveMetricCombo() const
{
    return m_metric;
}

QString MinimizerContainerItem::currentNormFunction() const
{
    return m_norm.currentValue();
}

void MinimizerContainerItem::setCurrentNormFunction(const QString& name)
{
    m_norm.setCurrentValue(name);
}

ComboProperty MinimizerContainerItem::normFunctionCombo() const
{
    return m_norm;
}

std::unique_ptr<IMinimizer> MinimizerContainerItem::createMinimizer() const
{
    return currentMinimizerItem()->createMinimizer();
}

std::unique_ptr<ObjectiveMetric> MinimizerContainerItem::createMetric() const
{
    return ObjectiveMetricUtils::createMetric(currentObjectiveMetric().toStdString(),
                                              currentNormFunction().toStdString());
}

void MinimizerContainerItem::writeTo(QXmlStreamWriter* writer) const
{
    // algorithm
    writer->writeStartElement(Tag::Algorithm);
    m_algorithm.writeTo(writer);
    writer->writeEndElement();

    // minimizer
    writer->writeStartElement(Tag::Minimizer);
    m_minimizer.writeTo(writer);
    writer->writeEndElement();

    // metric
    writer->writeStartElement(Tag::Metric);
    m_metric.writeTo(writer);
    writer->writeEndElement();

    // norm
    writer->writeStartElement(Tag::Norm);
    m_norm.writeTo(writer);
    writer->writeEndElement();

    // Minuit
    writer->writeStartElement(Tag::MinuitMinimizer);
    m_MinuitMinimizer->writeTo(writer);
    writer->writeEndElement();

    // GSLMulti
    writer->writeStartElement(Tag::GSLMultiMinimizer);
    m_GSLMultiMinimizer->writeTo(writer);
    writer->writeEndElement();

    // Genetic
    writer->writeStartElement(Tag::GeneticMinimizer);
    m_GeneticMinimizer->writeTo(writer);
    writer->writeEndElement();

    // SimAn
    writer->writeStartElement(Tag::SimAnMinimizer);
    m_SimAnMinimizer->writeTo(writer);
    writer->writeEndElement();

    // GSLLMA
    writer->writeStartElement(Tag::GSLLMAMinimizer);
    m_GSLLMAMinimizer->writeTo(writer);
    writer->writeEndElement();
}

void MinimizerContainerItem::readFrom(QXmlStreamReader* reader)
{
    while (reader->readNextStartElement()) {

        // algorithm
        if (reader->name() == Tag::Algorithm) {
            m_algorithm.readFrom(reader);
            GUI::Session::XML::gotoEndElementOfTag(reader, Tag::Algorithm);

            // minimizer
        } else if (reader->name() == Tag::Minimizer) {
            m_minimizer.readFrom(reader);
            GUI::Session::XML::gotoEndElementOfTag(reader, Tag::Minimizer);

            // metric
        } else if (reader->name() == Tag::Metric) {
            m_metric.readFrom(reader);
            GUI::Session::XML::gotoEndElementOfTag(reader, Tag::Metric);

            // norm
        } else if (reader->name() == Tag::Norm) {
            m_norm.readFrom(reader);
            GUI::Session::XML::gotoEndElementOfTag(reader, Tag::Norm);

            // Minuit
        } else if (reader->name() == Tag::MinuitMinimizer) {
            m_MinuitMinimizer->readFrom(reader);
            GUI::Session::XML::gotoEndElementOfTag(reader, Tag::MinuitMinimizer);

            // GSLMulti
        } else if (reader->name() == Tag::GSLMultiMinimizer) {
            m_GSLMultiMinimizer->readFrom(reader);
            GUI::Session::XML::gotoEndElementOfTag(reader, Tag::GSLMultiMinimizer);

            // Genetic
        } else if (reader->name() == Tag::GeneticMinimizer) {
            m_GeneticMinimizer->readFrom(reader);
            GUI::Session::XML::gotoEndElementOfTag(reader, Tag::GeneticMinimizer);

            // SimAn
        } else if (reader->name() == Tag::SimAnMinimizer) {
            m_SimAnMinimizer->readFrom(reader);
            GUI::Session::XML::gotoEndElementOfTag(reader, Tag::SimAnMinimizer);

            // GSLLMA
        } else if (reader->name() == Tag::GSLLMAMinimizer) {
            m_GSLLMAMinimizer->readFrom(reader);
            GUI::Session::XML::gotoEndElementOfTag(reader, Tag::GSLLMAMinimizer);

        } else
            reader->skipCurrentElement();
    }
}

// ----------------------------------------------------------------------------

MinuitMinimizerItem::MinuitMinimizerItem()
    : MinimizerItem()
    , m_strategy(1)
    , m_errorDef(1.0)
    , m_tolerance(0.01)
    , m_precision(-1.0)
    , m_maxFuncCalls(0)
{
    QString minimizer_name = minimizerTypeToName(MinimizerType::Minuit2);
    m_algorithm = MinimizerItemCatalog::algorithmCombo(minimizer_name);
}

MinuitMinimizerItem::~MinuitMinimizerItem() = default;

QString MinuitMinimizerItem::currentAlgorithm() const
{
    return m_algorithm.currentValue();
}

void MinuitMinimizerItem::setCurrentAlgorithm(const QString& name)
{
    m_algorithm.setCurrentValue(name);
}

int MinuitMinimizerItem::strategy() const
{
    return m_strategy;
}

void MinuitMinimizerItem::setStrategy(int value)
{
    m_strategy = value;
}

double MinuitMinimizerItem::errorDefinition() const
{
    return m_errorDef;
}

void MinuitMinimizerItem::setErrorDefinition(double value)
{
    m_errorDef = value;
}

double MinuitMinimizerItem::tolerance() const
{
    return m_tolerance;
}

void MinuitMinimizerItem::setTolerance(double value)
{
    m_tolerance = value;
}

double MinuitMinimizerItem::precision() const
{
    return m_precision;
}

void MinuitMinimizerItem::setPrecision(double value)
{
    m_precision = value;
}

int MinuitMinimizerItem::maxFuncCalls() const
{
    return m_maxFuncCalls;
}

void MinuitMinimizerItem::setMaxFuncCalls(int value)
{
    m_maxFuncCalls = value;
}

std::unique_ptr<IMinimizer> MinuitMinimizerItem::createMinimizer() const
{
    auto* domainMinimizer = new Minuit2Minimizer(currentAlgorithm().toStdString());
    domainMinimizer->setStrategy(strategy());
    domainMinimizer->setErrorDefinition(errorDefinition());
    domainMinimizer->setTolerance(tolerance());
    domainMinimizer->setPrecision(precision());
    domainMinimizer->setMaxFunctionCalls(maxFuncCalls());

    return std::unique_ptr<IMinimizer>(domainMinimizer);
}

void MinuitMinimizerItem::writeTo(QXmlStreamWriter* writer) const
{
    // algorithm
    writer->writeStartElement(Tag::Algorithm);
    m_algorithm.writeTo(writer);
    writer->writeEndElement();

    // strategy
    writer->writeStartElement(Tag::Strategy);
    GUI::Session::XML::writeAttribute(writer, Attrib::value, m_strategy);
    writer->writeEndElement();

    // error definition
    writer->writeStartElement(Tag::ErrorDef);
    GUI::Session::XML::writeAttribute(writer, Attrib::value, m_errorDef);
    writer->writeEndElement();

    // tolerance
    writer->writeStartElement(Tag::Tolerance);
    GUI::Session::XML::writeAttribute(writer, Attrib::value, m_tolerance);
    writer->writeEndElement();

    // precision
    writer->writeStartElement(Tag::Precision);
    GUI::Session::XML::writeAttribute(writer, Attrib::value, m_precision);
    writer->writeEndElement();

    // max function calls
    writer->writeStartElement(Tag::MaxFuncCalls);
    GUI::Session::XML::writeAttribute(writer, Attrib::value, m_maxFuncCalls);
    writer->writeEndElement();
}

void MinuitMinimizerItem::readFrom(QXmlStreamReader* reader)
{
    while (reader->readNextStartElement()) {

        // algorithm
        if (reader->name() == Tag::Algorithm) {
            m_algorithm.readFrom(reader);
            GUI::Session::XML::gotoEndElementOfTag(reader, Tag::Algorithm);

            // strategy
        } else if (reader->name() == Tag::Strategy) {
            GUI::Session::XML::readAttribute(reader, Attrib::value, &m_strategy);
            GUI::Session::XML::gotoEndElementOfTag(reader, Tag::Strategy);

            // error definition
        } else if (reader->name() == Tag::ErrorDef) {
            GUI::Session::XML::readAttribute(reader, Attrib::value, &m_errorDef);
            GUI::Session::XML::gotoEndElementOfTag(reader, Tag::ErrorDef);

            // tolerance
        } else if (reader->name() == Tag::Tolerance) {
            GUI::Session::XML::readAttribute(reader, Attrib::value, &m_tolerance);
            GUI::Session::XML::gotoEndElementOfTag(reader, Tag::Tolerance);

            // precision
        } else if (reader->name() == Tag::Precision) {
            GUI::Session::XML::readAttribute(reader, Attrib::value, &m_precision);
            GUI::Session::XML::gotoEndElementOfTag(reader, Tag::Precision);

            // max function calls
        } else if (reader->name() == Tag::MaxFuncCalls) {
            GUI::Session::XML::readAttribute(reader, Attrib::value, &m_maxFuncCalls);
            GUI::Session::XML::gotoEndElementOfTag(reader, Tag::MaxFuncCalls);

        } else
            reader->skipCurrentElement();
    }
}

// ----------------------------------------------------------------------------

GSLMultiMinimizerItem::GSLMultiMinimizerItem()
    : MinimizerItem()
    , m_maxIterations(0)
{
    QString minimizer_name = minimizerTypeToName(MinimizerType::GSLMultiMin);
    m_algorithm = MinimizerItemCatalog::algorithmCombo(minimizer_name);
}

GSLMultiMinimizerItem::~GSLMultiMinimizerItem() = default;

QString GSLMultiMinimizerItem::currentAlgorithm() const
{
    return m_algorithm.currentValue();
}

void GSLMultiMinimizerItem::setCurrentAlgorithm(const QString& name)
{
    m_algorithm.setCurrentValue(name);
}

int GSLMultiMinimizerItem::maxIterations() const
{
    return m_maxIterations;
}

void GSLMultiMinimizerItem::setMaxIterations(int value)
{
    m_maxIterations = value;
}

std::unique_ptr<IMinimizer> GSLMultiMinimizerItem::createMinimizer() const
{
    auto* domainMinimizer = new GSLMultiMinimizer(currentAlgorithm().toStdString());
    domainMinimizer->setMaxIterations(maxIterations());
    return std::unique_ptr<IMinimizer>(domainMinimizer);
}

void GSLMultiMinimizerItem::writeTo(QXmlStreamWriter* writer) const
{
    // algorithm
    writer->writeStartElement(Tag::Algorithm);
    m_algorithm.writeTo(writer);
    writer->writeEndElement();

    // max iterations
    writer->writeStartElement(Tag::MaxIterations);
    GUI::Session::XML::writeAttribute(writer, Attrib::value, m_maxIterations);
    writer->writeEndElement();
}

void GSLMultiMinimizerItem::readFrom(QXmlStreamReader* reader)
{
    while (reader->readNextStartElement()) {

        // algorithm
        if (reader->name() == Tag::Algorithm) {
            m_algorithm.readFrom(reader);
            GUI::Session::XML::gotoEndElementOfTag(reader, Tag::Algorithm);

            // max iterations
        } else if (reader->name() == Tag::MaxIterations) {
            GUI::Session::XML::readAttribute(reader, Attrib::value, &m_maxIterations);
            GUI::Session::XML::gotoEndElementOfTag(reader, Tag::MaxIterations);

        } else
            reader->skipCurrentElement();
    }
}

// ----------------------------------------------------------------------------

GeneticMinimizerItem::GeneticMinimizerItem()
    : MinimizerItem()
    , m_tolerance(0.01)
    , m_maxIterations(3)
    , m_populationSize(300)
    , m_randomSeed(0)
{
}

GeneticMinimizerItem::~GeneticMinimizerItem() = default;

double GeneticMinimizerItem::tolerance() const
{
    return m_tolerance;
}

void GeneticMinimizerItem::setTolerance(double value)
{
    m_tolerance = value;
}

int GeneticMinimizerItem::maxIterations() const
{
    return m_maxIterations;
}

void GeneticMinimizerItem::setMaxIterations(int value)
{
    m_maxIterations = value;
}

int GeneticMinimizerItem::populationSize() const
{
    return m_populationSize;
}

void GeneticMinimizerItem::setPopulationSize(int value)
{
    m_populationSize = value;
}

int GeneticMinimizerItem::randomSeed() const
{
    return m_randomSeed;
}

void GeneticMinimizerItem::setRandomSeed(int value)
{
    m_randomSeed = value;
}

std::unique_ptr<IMinimizer> GeneticMinimizerItem::createMinimizer() const
{
    auto* domainMinimizer = new GeneticMinimizer();
    domainMinimizer->setTolerance(tolerance());
    domainMinimizer->setMaxIterations(maxIterations());
    domainMinimizer->setPopulationSize(populationSize());
    domainMinimizer->setRandomSeed(randomSeed());
    return std::unique_ptr<IMinimizer>(domainMinimizer);
}

void GeneticMinimizerItem::writeTo(QXmlStreamWriter* writer) const
{
    // tolerance
    writer->writeStartElement(Tag::Tolerance);
    GUI::Session::XML::writeAttribute(writer, Attrib::value, m_tolerance);
    writer->writeEndElement();

    // max iterations
    writer->writeStartElement(Tag::MaxIterations);
    GUI::Session::XML::writeAttribute(writer, Attrib::value, m_maxIterations);
    writer->writeEndElement();

    // population size
    writer->writeStartElement(Tag::PopulationSize);
    GUI::Session::XML::writeAttribute(writer, Attrib::value, m_populationSize);
    writer->writeEndElement();

    // random seed
    writer->writeStartElement(Tag::RandomSeed);
    GUI::Session::XML::writeAttribute(writer, Attrib::value, m_randomSeed);
    writer->writeEndElement();
}

void GeneticMinimizerItem::readFrom(QXmlStreamReader* reader)
{
    while (reader->readNextStartElement()) {

        // tolerance
        if (reader->name() == Tag::Tolerance) {
            GUI::Session::XML::readAttribute(reader, Attrib::value, &m_tolerance);
            GUI::Session::XML::gotoEndElementOfTag(reader, Tag::Tolerance);

            // max iterations
        } else if (reader->name() == Tag::MaxIterations) {
            GUI::Session::XML::readAttribute(reader, Attrib::value, &m_maxIterations);
            GUI::Session::XML::gotoEndElementOfTag(reader, Tag::MaxIterations);

            // population size
        } else if (reader->name() == Tag::PopulationSize) {
            GUI::Session::XML::readAttribute(reader, Attrib::value, &m_populationSize);
            GUI::Session::XML::gotoEndElementOfTag(reader, Tag::PopulationSize);

            // random seed
        } else if (reader->name() == Tag::RandomSeed) {
            GUI::Session::XML::readAttribute(reader, Attrib::value, &m_randomSeed);
            GUI::Session::XML::gotoEndElementOfTag(reader, Tag::RandomSeed);

        } else
            reader->skipCurrentElement();
    }
}

// ----------------------------------------------------------------------------

SimAnMinimizerItem::SimAnMinimizerItem()
    : MinimizerItem()
    , m_maxIterations(100)
    , m_iterationsAtTemp(10)
    , m_stepSize(1.0)
    , m_boltzmann_K(1.0)
    , m_boltzmann_T_init(50.0)
    , m_boltzmann_Mu(1.05)
    , m_boltzmann_T_min(0.1)
{
}

SimAnMinimizerItem::~SimAnMinimizerItem() = default;

int SimAnMinimizerItem::maxIterations() const
{
    return m_maxIterations;
}

void SimAnMinimizerItem::setMaxIterations(int value)
{
    m_maxIterations = value;
}

int SimAnMinimizerItem::iterationsAtEachTemp() const
{
    return m_iterationsAtTemp;
}

void SimAnMinimizerItem::setIterationsAtEachTemp(int value)
{
    m_iterationsAtTemp = value;
}

double SimAnMinimizerItem::stepSize() const
{
    return m_stepSize;
}

void SimAnMinimizerItem::setStepSize(double value)
{
    m_stepSize = value;
}

double SimAnMinimizerItem::boltzmanK() const
{
    return m_boltzmann_K;
}

void SimAnMinimizerItem::setBoltzmanK(double value)
{
    m_boltzmann_K = value;
}

double SimAnMinimizerItem::boltzmanInitT() const
{
    return m_boltzmann_T_init;
}

void SimAnMinimizerItem::setBoltzmanInitT(double value)
{
    m_boltzmann_T_init = value;
}

double SimAnMinimizerItem::boltzmanMu() const
{
    return m_boltzmann_Mu;
}

void SimAnMinimizerItem::setBoltzmanMu(double value)
{
    m_boltzmann_Mu = value;
}

double SimAnMinimizerItem::boltzmanMinT() const
{
    return m_boltzmann_T_min;
}

void SimAnMinimizerItem::setBoltzmanMinT(double value)
{
    m_boltzmann_T_min = value;
}

std::unique_ptr<IMinimizer> SimAnMinimizerItem::createMinimizer() const
{
    auto* domainMinimizer = new SimAnMinimizer();
    domainMinimizer->setMaxIterations(maxIterations());
    domainMinimizer->setIterationsAtEachTemp(iterationsAtEachTemp());
    domainMinimizer->setStepSize(stepSize());
    domainMinimizer->setBoltzmannK(boltzmanK());
    domainMinimizer->setBoltzmannInitialTemp(boltzmanInitT());
    domainMinimizer->setBoltzmannMu(boltzmanMu());
    domainMinimizer->setBoltzmannMinTemp(boltzmanMinT());
    return std::unique_ptr<IMinimizer>(domainMinimizer);
}

void SimAnMinimizerItem::writeTo(QXmlStreamWriter* writer) const
{
    // max iterations
    writer->writeStartElement(Tag::MaxIterations);
    GUI::Session::XML::writeAttribute(writer, Attrib::value, m_maxIterations);
    writer->writeEndElement();

    // iterations at each temperature
    writer->writeStartElement(Tag::IterationsAtTemp);
    GUI::Session::XML::writeAttribute(writer, Attrib::value, m_iterationsAtTemp);
    writer->writeEndElement();

    // step size
    writer->writeStartElement(Tag::StepSize);
    GUI::Session::XML::writeAttribute(writer, Attrib::value, m_stepSize);
    writer->writeEndElement();

    // k
    writer->writeStartElement(Tag::BoltzmannK);
    GUI::Session::XML::writeAttribute(writer, Attrib::value, m_boltzmann_K);
    writer->writeEndElement();

    // T init
    writer->writeStartElement(Tag::Tinit);
    GUI::Session::XML::writeAttribute(writer, Attrib::value, m_boltzmann_T_init);
    writer->writeEndElement();

    // mu
    writer->writeStartElement(Tag::BoltzmannMu);
    GUI::Session::XML::writeAttribute(writer, Attrib::value, m_boltzmann_Mu);
    writer->writeEndElement();

    // T min
    writer->writeStartElement(Tag::Tmin);
    GUI::Session::XML::writeAttribute(writer, Attrib::value, m_boltzmann_T_min);
    writer->writeEndElement();
}

void SimAnMinimizerItem::readFrom(QXmlStreamReader* reader)
{
    while (reader->readNextStartElement()) {

        // max iterations
        if (reader->name() == Tag::MaxIterations) {
            GUI::Session::XML::readAttribute(reader, Attrib::value, &m_maxIterations);
            GUI::Session::XML::gotoEndElementOfTag(reader, Tag::MaxIterations);

            // iterations at each temperature
        } else if (reader->name() == Tag::IterationsAtTemp) {
            GUI::Session::XML::readAttribute(reader, Attrib::value, &m_iterationsAtTemp);
            GUI::Session::XML::gotoEndElementOfTag(reader, Tag::IterationsAtTemp);

            // step size
        } else if (reader->name() == Tag::StepSize) {
            GUI::Session::XML::readAttribute(reader, Attrib::value, &m_stepSize);
            GUI::Session::XML::gotoEndElementOfTag(reader, Tag::StepSize);

            // k
        } else if (reader->name() == Tag::BoltzmannK) {
            GUI::Session::XML::readAttribute(reader, Attrib::value, &m_boltzmann_K);
            GUI::Session::XML::gotoEndElementOfTag(reader, Tag::BoltzmannK);

            // T init
        } else if (reader->name() == Tag::Tinit) {
            GUI::Session::XML::readAttribute(reader, Attrib::value, &m_boltzmann_T_init);
            GUI::Session::XML::gotoEndElementOfTag(reader, Tag::Tinit);

            // mu
        } else if (reader->name() == Tag::BoltzmannMu) {
            GUI::Session::XML::readAttribute(reader, Attrib::value, &m_boltzmann_Mu);
            GUI::Session::XML::gotoEndElementOfTag(reader, Tag::BoltzmannMu);

            // T min
        } else if (reader->name() == Tag::Tmin) {
            GUI::Session::XML::readAttribute(reader, Attrib::value, &m_boltzmann_T_min);
            GUI::Session::XML::gotoEndElementOfTag(reader, Tag::Tmin);

        } else
            reader->skipCurrentElement();
    }
}

// ----------------------------------------------------------------------------

GSLLMAMinimizerItem::GSLLMAMinimizerItem()
    : MinimizerItem()
    , m_tolerance(0.01)
    , m_maxIterations(0)
{
}

GSLLMAMinimizerItem::~GSLLMAMinimizerItem() = default;

double GSLLMAMinimizerItem::tolerance() const
{
    return m_tolerance;
}

void GSLLMAMinimizerItem::setTolerance(double value)
{
    m_tolerance = value;
}

int GSLLMAMinimizerItem::maxIterations() const
{
    return m_maxIterations;
}

void GSLLMAMinimizerItem::setMaxIterations(int value)
{
    m_maxIterations = value;
}

std::unique_ptr<IMinimizer> GSLLMAMinimizerItem::createMinimizer() const
{
    auto* domainMinimizer = new GSLLevenbergMarquardtMinimizer();
    domainMinimizer->setTolerance(tolerance());
    domainMinimizer->setMaxIterations(maxIterations());
    return std::unique_ptr<IMinimizer>(domainMinimizer);
}

void GSLLMAMinimizerItem::writeTo(QXmlStreamWriter* writer) const
{
    // tolerance
    writer->writeStartElement(Tag::Tolerance);
    GUI::Session::XML::writeAttribute(writer, Attrib::value, m_tolerance);
    writer->writeEndElement();

    // max iterations
    writer->writeStartElement(Tag::MaxIterations);
    GUI::Session::XML::writeAttribute(writer, Attrib::value, m_maxIterations);
    writer->writeEndElement();
}

void GSLLMAMinimizerItem::readFrom(QXmlStreamReader* reader)
{
    while (reader->readNextStartElement()) {

        // tolerance
        if (reader->name() == Tag::Tolerance) {
            GUI::Session::XML::readAttribute(reader, Attrib::value, &m_tolerance);
            GUI::Session::XML::gotoEndElementOfTag(reader, Tag::Tolerance);

            // max iterations
        } else if (reader->name() == Tag::MaxIterations) {
            GUI::Session::XML::readAttribute(reader, Attrib::value, &m_maxIterations);
            GUI::Session::XML::gotoEndElementOfTag(reader, Tag::MaxIterations);

        } else
            reader->skipCurrentElement();
    }
}
