//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      GUI/Model/Model/ApplicationModels.cpp
//! @brief     Defines class holding all application models
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2018
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#include "GUI/Model/Model/ApplicationModels.h"
#include "GUI/Support/Data/SimulationOptionsItem.h"
#include "GUI/Util/DeserializationException.h"
#include <QXmlStreamWriter>

namespace {

namespace Tag {

const QString JobModel("JobModel");
const QString RealModel("RealModel");

} // namespace Tag

} // namespace

ApplicationModels::ApplicationModels()
{
    //! creates and initializes models, order is important
    m_realDataModel = std::make_unique<RealModel>(this);
    m_jobModel = std::make_unique<JobModel>(this);

    connectModel(m_realDataModel.get());
    connectModel(m_jobModel.get());
}

RealModel* ApplicationModels::realModel() const
{
    return m_realDataModel.get();
}

JobModel* ApplicationModels::jobModel() const
{
    return m_jobModel.get();
}

//! Writes all model in file one by one

void ApplicationModels::writeTo(QXmlStreamWriter* writer)
{
    writer->writeStartElement(Tag::RealModel);
    m_realDataModel->writeTo(writer);
    writer->writeEndElement();

    writer->writeStartElement(Tag::JobModel);
    m_jobModel->writeTo(writer);
    writer->writeEndElement();
}

void ApplicationModels::readFrom(QXmlStreamReader* reader, MessageService* messageService)
{
    try {
        if (reader->name() == Tag::RealModel)
            m_realDataModel->readFrom(reader, messageService);
        else if (reader->name() == Tag::JobModel)
            m_jobModel->readFrom(reader, messageService);
    } catch (DeserializationException& ex) {
        m_realDataModel->clear();
        m_jobModel->clear();
        reader->raiseError(ex.text());
    }
}


QVector<DataItem*> ApplicationModels::dataItems() const
{
    ASSERT(m_realDataModel && m_jobModel);
    return QVector<DataItem*>() << m_realDataModel->dataItems() << m_jobModel->dataItems();
}

void ApplicationModels::connectModel(SessionModel* model) const
{
    if (!model)
        return;
    connect(model, &SessionModel::dataChanged, this, &ApplicationModels::modelChanged,
            Qt::UniqueConnection);
    connect(model, &SessionModel::rowsRemoved, this, &ApplicationModels::modelChanged,
            Qt::UniqueConnection);
    connect(model, &SessionModel::rowsInserted, this, &ApplicationModels::modelChanged,
            Qt::UniqueConnection);
}
