//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      GUI/Model/Sample/CompoundItem.cpp
//! @brief     Implements class CompoundItem
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2021
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#include "GUI/Model/Sample/CompoundItem.h"
#include "GUI/Model/CatSample/ItemWithParticlesCatalog.h"
#include "GUI/Model/CatSample/RotationItemCatalog.h"
#include "GUI/Model/Sample/CoreAndShellItem.h"
#include "GUI/Model/Sample/MesocrystalItem.h"
#include "GUI/Model/Sample/ParticleItem.h"
#include "GUI/Support/XML/Serialize.h"
#include "Sample/Particle/CoreAndShell.h"
#include "Sample/Particle/Mesocrystal.h"
#include "Sample/Particle/Particle.h"
#include "Sample/Scattering/Rotations.h"

namespace {

namespace Tag {

const QString Abundance("Abundance");
const QString Position("Position");
const QString Particles("Particles");
const QString Rotation("Rotation");

} // namespace Tag

const QString abundance_tooltip = "Proportion of this type of particles normalized to the \n"
                                  "total number of particles in the layout";

const QString position_tooltip = "Relative position of the particle's reference point \n"
                                 "in the coordinate system of the parent (nm)";

} // namespace


CompoundItem::CompoundItem(const MaterialItems* materials)
    : ItemWithParticles(abundance_tooltip, position_tooltip)
    , m_materials(materials)
{
    ASSERT(m_materials);
}

void CompoundItem::serialize(Streamer& s)
{
    s.assertVersion(0);
    Serialize::rwProperty(s, Tag::Abundance, m_abundance);
    m_position.rwProperty(s, Tag::Position);
    m_rotation.rwSelected(s, Tag::Rotation);
    Serialize::rwCatalogized<ItemWithParticlesCatalog>(s, Tag::Particles, m_particles, m_materials);
}

std::unique_ptr<Compound> CompoundItem::createCompound() const
{
    auto P_composition = std::make_unique<Compound>();
    P_composition->setAbundance(abundance());
    for (auto* p : m_particles) {
        if (auto* particleItem = dynamic_cast<ParticleItem*>(p)) {
            if (auto P_particle = particleItem->createParticle())
                P_composition->addParticle(*P_particle);
        } else if (auto* coreShellItem = dynamic_cast<CoreAndShellItem*>(p)) {
            if (auto P_particle_coreshell = coreShellItem->createCoreAndShell())
                P_composition->addParticle(*P_particle_coreshell);
        } else if (auto* compositionItem = dynamic_cast<CompoundItem*>(p)) {
            if (auto P_child_composition = compositionItem->createCompound())
                P_composition->addParticle(*P_child_composition);
        } else if (auto* mesocrystalItem = dynamic_cast<MesocrystalItem*>(p)) {
            if (auto P_meso = mesocrystalItem->createMesocrystal())
                P_composition->addParticle(*P_meso);
        }
    }

    P_composition->setParticlePosition(position());
    if (const auto r = createRotation(); r && !r->isIdentity())
        P_composition->setRotation(*r);

    return P_composition;
}

QVector<ItemWithParticles*> CompoundItem::particles() const
{
    return m_particles;
}

void CompoundItem::addParticle(ItemWithParticles* particle)
{
    m_particles << particle;
}

void CompoundItem::removeParticle(ItemWithParticles* particle)
{
    m_particles.removeAll(particle);
    delete particle;
}

QVector<ItemWithParticles*> CompoundItem::containedItemsWithParticles() const
{
    QVector<ItemWithParticles*> result;
    for (auto* particle : m_particles)
        result << particle << particle->containedItemsWithParticles();
    return result;
}
