//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      GUI/Model/Sample/Lattice2DItems.cpp
//! @brief     Implements classes Lattice2DItems
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2018
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#include "GUI/Model/Sample/Lattice2DItems.h"
#include "Base/Const/Units.h"
#include "Base/Util/Assert.h"
#include "GUI/Support/XML/Serialize.h"
#include "Sample/Lattice/Lattice2D.h"

namespace {
namespace Tag {

const QString LatticeRotationAngle("LatticeRotationAngle");
const QString Length("Length");
const QString Length1("Length1");
const QString Length2("Length2");
const QString Angle("Angle");

} // namespace Tag
} // namespace

Lattice2DItem::Lattice2DItem()
{
    m_latticeRotationAngle.init(
        "Xi", "Rotation of lattice with respect to x-axis of reference frame (beam direction)", 0.0,
        Unit::degree, "xi");
}

double Lattice2DItem::unitCellArea() const
{
    return createLattice()->unitCellArea();
}

// --------------------------------------------------------------------------------------------- //

BasicLattice2DItem::BasicLattice2DItem()
{
    m_length1.init("LatticeLength1", "Length of first lattice vector", 20.0, Unit::nanometer,
                   "len1");
    m_length2.init("LatticeLength2", "Length of second lattice vector", 20.0, Unit::nanometer,
                   "len2");
    m_angle.init("Angle", "Angle between lattice vectors", 90.0, Unit::degree, "angle");
}

std::unique_ptr<Lattice2D> BasicLattice2DItem::createLattice() const
{
    return std::make_unique<BasicLattice2D>(m_length1.value(), m_length2.value(),
                                            Units::deg2rad(m_angle.value()),
                                            Units::deg2rad(m_latticeRotationAngle.value()));
}

void BasicLattice2DItem::serialize(Streamer& s)
{
    s.assertVersion(0);
    Serialize::rwProperty(s, Tag::LatticeRotationAngle, m_latticeRotationAngle);
    Serialize::rwProperty(s, Tag::Length1, m_length1);
    Serialize::rwProperty(s, Tag::Length2, m_length2);
    Serialize::rwProperty(s, Tag::Angle, m_angle);
}

// --------------------------------------------------------------------------------------------- //

SquareLattice2DItem::SquareLattice2DItem()
{
    m_length.init("LatticeLength", "Length of first and second lattice vectors", 20.0,
                  Unit::nanometer, "len");
}

std::unique_ptr<Lattice2D> SquareLattice2DItem::createLattice() const
{
    return std::make_unique<SquareLattice2D>(m_length,
                                             Units::deg2rad(m_latticeRotationAngle.value()));
}

void SquareLattice2DItem::serialize(Streamer& s)
{
    s.assertVersion(0);
    Serialize::rwProperty(s, Tag::LatticeRotationAngle, m_latticeRotationAngle);
    Serialize::rwProperty(s, Tag::Length, m_length);
}

// --------------------------------------------------------------------------------------------- //

HexagonalLattice2DItem::HexagonalLattice2DItem()
{
    m_length.init("LatticeLength", "Length of first and second lattice vectors", 20.0,
                  Unit::nanometer, "len");
}

std::unique_ptr<Lattice2D> HexagonalLattice2DItem::createLattice() const
{
    return std::make_unique<HexagonalLattice2D>(m_length.value(),
                                                Units::deg2rad(m_latticeRotationAngle.value()));
}

void HexagonalLattice2DItem::serialize(Streamer& s)
{
    s.assertVersion(0);
    Serialize::rwProperty(s, Tag::LatticeRotationAngle, m_latticeRotationAngle);
    Serialize::rwProperty(s, Tag::Length, m_length);
}
