//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      Sample/HardParticle/Bipyramid4.cpp
//! @brief     Implements class Bipyramid4.
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2018
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#include "Sample/HardParticle/Bipyramid4.h"
#include "Base/Math/Constants.h"
#include "Base/Math/Functions.h"
#include "Sample/HardParticle/Pyramid4.h"

const ff::Topology Bipyramid4::topology = {{{{3, 2, 1, 0}, true},
                                            {{0, 1, 5, 4}, false},
                                            {{1, 2, 6, 5}, false},
                                            {{2, 3, 7, 6}, false},
                                            {{3, 0, 4, 7}, false},
                                            {{4, 5, 9, 8}, false},
                                            {{5, 6, 10, 9}, false},
                                            {{6, 7, 11, 10}, false},
                                            {{7, 4, 8, 11}, false},
                                            {{8, 9, 10, 11}, true}},
                                           false};

Bipyramid4::Bipyramid4(const std::vector<double> P)
    : IFormFactorPolyhedron(P)
    , m_length(m_P[0])
    , m_height(m_P[1])
    , m_height_ratio(m_P[2])
    , m_alpha(m_P[3])
{
    validateOrThrow();
}

Bipyramid4::Bipyramid4(double length, double height, double height_ratio, double alpha)
    : Bipyramid4(std::vector<double>{length, height, height_ratio, alpha})
{
}

std::string Bipyramid4::validate() const
{
    std::vector<std::string> errs;
    requestGt0(errs, m_length, "length");
    requestGt0(errs, m_height, "height");
    requestGt0(errs, m_height_ratio, "height_ratio");

    const double cot_alpha = Math::cot(m_alpha);
    if (m_alpha <= 0 || !std::isfinite(cot_alpha) || cot_alpha < 0)
        errs.push_back("pyramid angle alpha=" + std::to_string(m_alpha) + " out of bounds");

    if (!errs.empty())
        return jointError(errs);

    const double x = m_height_ratio;
    const double r = cot_alpha * 2 * m_height / m_length;
    if (std::max(1., x) * r > 1)
        errs.push_back(
            "parameters violate condition 2*height <= length*tan(alpha)*min(1,1/height_ratio)");
    if (!errs.empty())
        return jointError(errs);

    const double a = m_length / 2 * (1 - r);
    const double b = m_length / 2;
    const double c = m_length / 2 * (1 - r * x);

    const double dzcom =
        m_height * ((x * x - 1) / 2 - 2 * r * (x * x * x - 1) / 3 + r * r * (x * x * x * x - 1) / 4)
        / ((x + 1) - r * (x * x + 1) + r * r * (x * x * x + 1) / 3);
    const double za = -dzcom - m_height;
    const double zb = -dzcom;
    const double zc = -dzcom + x * m_height;

    setPolyhedron(topology, za,
                  {// base:
                   {-a, -a, za},
                   {a, -a, za},
                   {a, a, za},
                   {-a, a, za},
                   // middle
                   {-b, -b, zb},
                   {b, -b, zb},
                   {b, b, zb},
                   {-b, b, zb},
                   // top
                   {-c, -c, zc},
                   {c, -c, zc},
                   {c, c, zc},
                   {-c, c, zc}});

    m_validated = true;
    return "";
}
