//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      Sample/HardParticle/Pyramid3.cpp
//! @brief     Implements class Pyramid3.
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2018
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#include "Sample/HardParticle/Pyramid3.h"
#include "Base/Math/Constants.h"
#include "Base/Math/Functions.h"

const ff::Topology Pyramid3::topology = {{{{2, 1, 0}, false},
                                          {{0, 1, 4, 3}, false},
                                          {{1, 2, 5, 4}, false},
                                          {{2, 0, 3, 5}, false},
                                          {{3, 4, 5}, false}},
                                         false};

Pyramid3::Pyramid3(const std::vector<double> P)
    : IFormFactorPolyhedron(P)
    , m_base_edge(m_P[0])
    , m_height(m_P[1])
    , m_alpha(m_P[2])
{
    validateOrThrow();
}

Pyramid3::Pyramid3(double base_edge, double height, double alpha)
    : Pyramid3(std::vector<double>{base_edge, height, alpha})
{
}

std::string Pyramid3::validate() const
{
    std::vector<std::string> errs;
    requestGt0(errs, m_base_edge, "base_edge");
    requestGt0(errs, m_height, "height");
    const double cot_alpha = Math::cot(m_alpha);
    if (m_alpha <= 0 || !std::isfinite(cot_alpha) || cot_alpha < 0)
        errs.push_back("pyramid angle alpha " + std::to_string(m_alpha) + " out of bounds");
    if (!errs.empty())
        return jointError(errs);

    double r = cot_alpha * 2 * sqrt(3.) * m_height / m_base_edge; // L(top)/L(base)
    if (r > 1)
        return jointError(
            {"parameters violate condition 2*sqrt(3)*ctg(alpha) <= base_edge/height"});

    double a = m_base_edge;
    double as = a / 2;
    double ac = a * (1. / sqrt(3) / 2);
    double ah = a * (1. / sqrt(3));
    double b = a * (1 - r);
    double bs = b / 2;
    double bc = b * (1. / sqrt(3) / 2);
    double bh = b * (1. / sqrt(3));

    // center of mass
    double zcom = m_height * (1. / 4) * (6 - 8 * r + 3 * r * r) / (3 - 3 * r + r * r);

    setPolyhedron(topology, -zcom,
                  {// base:
                   {-ac, as, -zcom},
                   {-ac, -as, -zcom},
                   {ah, 0., -zcom},
                   // top:
                   {-bc, bs, m_height - zcom},
                   {-bc, -bs, m_height - zcom},
                   {bh, 0., m_height - zcom}});

    m_validated = true;
    return "";
}
