//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      Sample/HardParticle/TruncatedCube.cpp
//! @brief     Implements class TruncatedCube.
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2018
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#include "Sample/HardParticle/TruncatedCube.h"

const ff::Topology TruncatedCube::topology = {{{{0, 1, 7, 6, 9, 10, 4, 3}, true},
                                               {{0, 2, 1}, false},
                                               {{3, 4, 5}, false},
                                               {{9, 11, 10}, false},
                                               {{6, 7, 8}, false},
                                               {{0, 3, 5, 17, 15, 12, 14, 2}, true},
                                               {{4, 10, 11, 23, 22, 16, 17, 5}, true},
                                               {{1, 2, 14, 13, 19, 20, 8, 7}, true},
                                               {{6, 8, 20, 18, 21, 23, 11, 9}, true},
                                               {{15, 17, 16}, false},
                                               {{12, 13, 14}, false},
                                               {{18, 20, 19}, false},
                                               {{21, 22, 23}, false},
                                               {{12, 15, 16, 22, 21, 18, 19, 13}, true}},
                                              true};

TruncatedCube::TruncatedCube(const std::vector<double> P)
    : IFormFactorPolyhedron(P)
    , m_length(m_P[0])
    , m_removed_length(m_P[1])
{
    validateOrThrow();
}

TruncatedCube::TruncatedCube(double length, double removed_length)
    : TruncatedCube(std::vector<double>{length, removed_length})
{
}

std::string TruncatedCube::validate() const
{
    std::vector<std::string> errs;
    requestGt0(errs, m_length, "length");
    requestGe0(errs, m_removed_length, "removed_length");
    if (!errs.empty())
        return jointError(errs);

    const double a = m_length / 2;
    const double c = a - m_removed_length;
    if (c < 0)
        return jointError({"parameters violate condition removed_length <= 0.5*length"});


    setPolyhedron(topology, -a,
                  {{-c, -a, -a}, {-a, -c, -a}, {-a, -a, -c}, {c, -a, -a}, {a, -c, -a}, {a, -a, -c},
                   {-c, a, -a},  {-a, c, -a},  {-a, a, -c},  {c, a, -a},  {a, c, -a},  {a, a, -c},
                   {-c, -a, a},  {-a, -c, a},  {-a, -a, c},  {c, -a, a},  {a, -c, a},  {a, -a, c},
                   {-c, a, a},   {-a, c, a},   {-a, a, c},   {c, a, a},   {a, c, a},   {a, a, c}});


    m_validated = true;
    return "";
}
